# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import unittest
import shutil
import processing
import tempfile, os
import debug
import logging

class TestCase(unittest.TestCase):

    def setUp(self):
        """docstring for setUp"""
        self.files = []
        self.directories = []
        self.processes = []
        if not hasattr(self, 'logging_is_configured'):
            debug.configure_logging(type(self).__name__, logging.INFO)
            self.logging_is_configured = True

    def tearDown(self):
        """docstring for tearDown"""
        for process in self.processes:
            process.terminate()
        for filepath in self.files:
            if os.path.isfile(filepath):
                os.unlink(filepath)
        for dirpath in self.directories:
            shutil.rmtree(dirpath)

    def tempdir(self):
        """docstring for tempdir"""
        dirpath = tempfile.mkdtemp()
        self.directories.append(dirpath)
        return dirpath

    def tempfile(self):
        """docstring for tempfile"""
        handle, filepath = tempfile.mkstemp()
        os.close(handle)
        self.files.append(filepath)
        return filepath

    def startProcess(self, target, **kwargs):
        process = processing.Process(target=target, **kwargs)
        self.processes.append(process)
        process.start()



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import os, sys

def parent(path):
    assert path
    return os.path.dirname(path)

def find_root(root_name):
    root = parent(os.path.abspath(__file__))
    while os.path.basename(root) != root_name:
        if root == '/':
            raise Exception('Project root not found!')
        root = parent(root)
    return root

def path(_path):
    """docstring for path"""
    return os.path.join(home, _path)

home = parent(find_root('tako'))
data = os.path.join(home, 'var', 'data')
log = os.path.join(home, 'var', 'log')

def setup():
    sys.path.append(home)



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import logging
import os
import simplejson as json

import configuration
from utils import debug
from utils import timestamper
from utils import httpserver

class CoordinatorServer(object):
    def __init__(self, coordinator_id, configuration_filepath):
        super(CoordinatorServer, self).__init__()
        self.id = coordinator_id
        self.configuration = None
        self.configuration_filepath = configuration_filepath
        self.__http_handlers = (
            ('/configuration', {'GET': self.configuration_GET}),
        )
        self.http_server = None
        self.reload_configuration()

    def reload_configuration(self):
        self.configuration = configuration.try_load_file(self.configuration_filepath)
        if not self.configuration:
            raise Exception('Failed to load configuration from file: "%s"', self.configuration_filepath)
        self.configuration.timestamp = timestamper.from_seconds(os.stat(self.configuration_filepath).st_mtime)
        self.coordinator = self.configuration.coordinators[self.id]
        logging.debug('timestamp: %s', self.configuration.timestamp)

    def configuration_GET(self, start_response, path, body, env):
        logging.info('%(REMOTE_ADDR)s [GET %(PATH_INFO)s %(SERVER_PROTOCOL)s]', env)
        start_response('200 OK', [
            ('content-type', 'application/json'),
            ('x-timestamp', str(self.configuration.timestamp)),
        ])
        return [json.dumps(self.configuration.representation())]

    def serve(self):
        logging.info('http://%s:%s/configuration', self.coordinator.address, self.coordinator.port)
        self.http_server = httpserver.HttpServer(listener=('', self.coordinator.port), handlers=self.__http_handlers)
        self.http_server.serve()



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import logging

from syncless import coio

import paths
paths.setup()

import configuration
from utils import timestamper
from utils import http

class CoordinatorClient(object):
    def __init__(self, coordinators=[], callbacks=None, interval=30):
        super(CoordinatorClient, self).__init__()
        self.coordinators = coordinators
        self.callbacks = callbacks
        self.interval = interval
        self.configuration_fetcher = None
        self.configuration = None

    def start(self):
        self.configuration_fetcher = coio.stackless.tasklet(self.__fetch_configurations)()

    def stop(self):
        if self.configuration_fetcher:
            self.configuration_fetcher.kill()
            self.configuration_fetcher = None

    def __notify(self):
        logging.debug('configuration: %s', self.configuration)
        for f in self.callbacks:
            f(self.configuration)

    def __fetch_configuration(self, coordinator):
        # logging.debug('coordinator: %s', coordinator)
        url = coordinator.configuration_url()
        body, info = http.fetch(url)
        if body:
            # logging.debug('Got representation: %s', body)
            new_timestamp = timestamper.try_loads(info.get('x-timestamp', None))
            if new_timestamp:
                new_configuration = configuration.try_load_json(body, timestamp=new_timestamp)
                return (new_configuration, coordinator)
        return (None, coordinator)

    def __set_configuration(self, new_configuration):
        self.configuration = new_configuration
        self.coordinators = new_configuration.coordinators.values()

    def __fetch_configurations(self):
        while True:
            logging.debug('coordinators: %s', self.coordinators)
            if self.coordinators:
                configurations = []
                for coordinator in self.coordinators:
                    configurations.append(self.__fetch_configuration(coordinator))
                configurations.sort()
                for new_configuration, source_coordinator in configurations:
                    if new_configuration and (not self.configuration or new_configuration.timestamp > self.configuration.timestamp):
                        self.__set_configuration(new_configuration)
                        self.__notify()
                        break
            coio.sleep(self.interval)




import urllib
import logging
import os
import email.utils

import paths
paths.setup()

from utils import debug
from utils import timestamper
from utils import httpserver

from client import Client, ValueNotAvailableException

class ProxyServer(object):
    """docstring for ProxyServer"""
    def __init__(self, proxy_id, address, coordinator_addresses, explicit_configuration, var_directory='var'):
        super(ProxyServer, self).__init__()
        self.id = proxy_id
        self.address = address
        self.__http_handlers = (
                ('/values/', {'GET':self.__values_GET, 'POST':self.__values_POST}),
                ('/stat/', {'GET':self.__stat_GET}),
        )
        configuration_cache_directory = os.path.join(var_directory, 'etc')
        name = 'proxyserver-%s' % self.id
        self.__client = Client(name, coordinator_addresses, explicit_configuration, configuration_cache_directory)

    def __get_timestamp(self, env):
        try:
            return timestamper.try_loads(env.get('HTTP_X_TIMESTAMP', None)) or timestamper.now()
        except ValueError:
            raise httpserver.BadRequest()

    def __quote(self, key):
        return urllib.quote_plus(key, safe='/&')

    def __unquote(self, path):
        return urllib.unquote_plus(path)

    def __values_GET(self, start_response, path, body, env):
        debug.log('path: %s', path)
        key = self.__unquote(path)
        try:
            timestamp, value = self.__client.get(key)
        except ValueNotAvailableException:
            start_response('503 Service Unavailable', [])
            return ['']

        if timestamp and value:
            start_response('200 OK', [
                    ('Content-Type', 'application/octet-stream'),
                    ('Last-Modified', email.utils.formatdate(timestamper.to_seconds(timestamp))),
                    ('X-Timestamp', str(timestamp)),
            ])
            return [value]
        else:
            start_response('404 Not Found', [])
            return ['']

    def __values_POST(self, start_response, path, body, env):
        debug.log("path: %s", path)
        key = self.__unquote(path)
        value = body.read()
        timestamp = self.__get_timestamp(env)
        try:
            new_timestamp = self.__client.set(key, timestamp, value)
        except ValueNotAvailableException:
            start_response('503 Service Unavailable', [])
            return ['']

        if new_timestamp:
            start_response('200 OK', [
                    ('Content-Type', 'application/octet-stream'),
                    ('X-Timestamp', str(new_timestamp)),
            ])
            return ['']
        else:
            start_response('503 Service Unavailable', [])
            return ['']

    def __stat_GET(self, start_response, path, body, env):
        debug.log('path: %s', path)
        key = self.__unquote(path)

        try:
            timestamp = self.__client.stat(key)
        except ValueNotAvailableException:
            start_response('503 Service Unavailable', [])
            return ['']

        if timestamp:
            start_response('200 OK', [
                    ('Content-Type', 'application/octet-stream'),
                    ('Last-Modified', email.utils.formatdate(timestamper.to_seconds(timestamp))),
                    ('X-Timestamp', timestamper.dumps(timestamp)),
            ])
            return [timestamper.dumps(timestamp)]
        else:
            start_response('404 Not Found', [])
            return ['']

    def serve(self):
        logging.info('Public HTTP API: %s:%s' % self.address)
        self.__client.connect()
        self.__http_server = httpserver.HttpServer(listener=self.address,
                                                 handlers=self.__http_handlers)
        self.__http_server.serve()



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

def try_float(s):
    try:
        return float(s)
    except ValueError:
        return None
    except TypeError:
        return None

def try_int(s):
    try:
        return int(s)
    except ValueError:
        return None
    except TypeError:
        return None



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-
import logging

import paths
paths.setup()

from utils import debug

from configurationcache import ConfigurationCache
from coordinatorclient import CoordinatorClient
from models import Coordinator


class ConfigurationController(object):
    """docstring for ConfigurationController"""
    def __init__(self, name, coordinator_addresses, explicit_configuration, configuration_cache_directory,
                 update_configuration_callback, update_interval=5*60):
        super(ConfigurationController, self).__init__()
        self.name = name
        coordinators = [Coordinator(None, address, port) for address, port in coordinator_addresses]
        self.__coordinator_client = CoordinatorClient(coordinators=coordinators,
                                                      callbacks=[self.__evaluate_new_configuration],
                                                      interval=update_interval)
        self.__configuration_cache = None
        self.__explicit_configuration = explicit_configuration
        self.__update_configuration_callback = None

        if configuration_cache_directory:
            debug.log('Persistent configuration cache enabled.')
            self.__configuration_cache = ConfigurationCache(configuration_cache_directory, self.name)
        else:
            debug.log('No configuration cache directory., persistent configuration cache disabled.')

        self.configuration = None
        if explicit_configuration:
            debug.log('Using explicit configuration')
            self.__evaluate_new_configuration(explicit_configuration)
        elif self.__configuration_cache:
            cached_configuration = self.__configuration_cache.get_configuration()
            if cached_configuration:
                debug.log('Using cached configuration')
                self.__evaluate_new_configuration(cached_configuration)

        self.__update_configuration_callback = update_configuration_callback

        if not self.configuration and not coordinator_addresses:
            raise Exception('No configuration available.')

    def __evaluate_new_configuration(self, new_configuration):
        if not self.configuration or new_configuration.timestamp > self.configuration.timestamp:
            if self.configuration:
                logging.info('New configuration: %s', new_configuration)
            self.configuration = new_configuration
            if self.__configuration_cache:
                self.__configuration_cache.cache_configuration(self.configuration)
            if self.__update_configuration_callback:
                self.__update_configuration_callback(new_configuration)

    def start(self):
        if self.configuration and self.__update_configuration_callback:
            self.__update_configuration_callback(self.configuration)
        self.__coordinator_client.start()

    def stop(self):
        self.__coordinator_client.stop()



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

from __future__ import with_statement

import yaml
import simplejson as json
import logging
import os

import paths
paths.setup()

from utils import debug
from utils import timestamper
from models import Coordinator, Deployment

def try_load_representation(representation, timestamp=timestamper.now()):
    try:
        logging.debug(representation)
        configuration = Configuration(representation, timestamp)
        logging.debug(configuration)
        logging.debug(configuration.representation())
        return configuration
    except ValidationError, e:
        logging.error('Configuration is not valid: %s', e)
        return None

def try_load_json(s, timestamp=timestamper.now()):
    try:
        representation = json.loads(s)
    except Exception, e:
        logging.error('Failed to parse JSON configuration: %s', e)
        return None
    return try_load_representation(representation, timestamp)

def try_load_file(filepath, timestamp=None):
    try:
        if not timestamp:
            timestamp = timestamper.from_seconds(os.path.getmtime(filepath))
        with open(filepath) as f:
            representation = yaml.load(f)
    except OSError, e:
        logging.error('Failed reading configuration file: %s', e)
        return None
    except IOError, e:
        logging.error('Failed reading configuration file: %s', e)
        return None
    except Exception, e:
        logging.error('Failed reading configuration file: %s', e)
        return None
    return try_load_representation(representation, timestamp)

def try_dump_file(filepath, configuration):
    logging.debug(filepath)
    try:
        dirpath = os.path.dirname(filepath)
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)
        with open(filepath, 'w+') as f:
            yaml.dump(configuration.representation(), f)
        return True
    except IOError, e:
        logging.error('Failed to write configuration to file: %s', e)
        return False
    except OSError, e:
        logging.error('Failed to write configuration to file: %s', e)
        return False

def validate_representation(representation):
    # TODO: Complete validation
    # TODO: Validate that node id's are not recycled/nodes are not changed
    try:
        assert 'active_deployment' in representation
        assert 'deployments' in representation
        assert len(representation['deployments']) > 0
        assert representation['active_deployment'] in representation['deployments']
        if 'target_deployment' in representation:
            assert representation['target_deployment'] in representation['deployments']
        for deployment_id, deployment in representation['deployments'].iteritems():
            assert 'buckets' in deployment
            for bucket_id, bucket in deployment['buckets'].iteritems():
                assert len(bucket) > 0
                for node_id, node in bucket.iteritems():
                    assert len(node) == 3
                    address, http_port, raw_port = node
                    assert type(address) == str, 'type(address) == %s != str' % type(address)
                    assert type(http_port) == int, 'type(http_port) == %s != int' % type(http_port)
                    assert type(raw_port) == int, 'type(raw_port) == %s != int' % type(raw_port)
        if 'master_coordinator' in representation:
            assert representation['master_coordinator'] in representation['coordinators']
        for coordinator_id, coordinator in representation.get('coordinators', {}).iteritems():
            assert len(coordinator) == 2
            address, port = coordinator
            assert type(address) == str, 'type(address) == %s != str' % type(address)
            assert type(port) == int, 'type(port) == %s != int' % type(port)
    except AssertionError:
        raise
        # raise ValidationError()

class ValidationError(Exception):
    def __init__(self, description=None):
        super(ValidationError, self).__init__()
        self.description = description

    def __str__(self):
        return 'ValidationError(%s)' % (self.description or '')

    def __repr__(self):
        return str(self)

class Configuration(object):
    def __init__(self, representation=None, timestamp=timestamper.now()):
        super(Configuration, self).__init__()
        self.timestamp = timestamp
        if representation:
            assert self.load(representation)

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "Configuration(%s)" % self.representation()

    def load(self, representation):
        validate_representation(representation)
        self.original_representation = representation
        self.deployments = dict((name, Deployment(name, deployment_representation)) for name, deployment_representation in representation.get('deployments', {}).iteritems())
        self.active_deployment_name = representation['active_deployment']
        self.active_deployment = self.deployments[self.active_deployment_name]
        self.target_deployment_name = representation.get('target_deployment', None)
        self.target_deployment = self.deployments[self.target_deployment_name] if self.target_deployment_name else None
        self.coordinators = dict((coordinator_id, Coordinator(coordinator_id, address, port)) for coordinator_id, (address, port) in representation.get('coordinators', {}).iteritems())
        self.master_coordinator_id = representation.get('master_coordinator', None)
        self.master_coordinator = self.coordinators.get(self.master_coordinator_id, None)
        return True

    def representation(self):
        spec = {
                'active_deployment': self.active_deployment_name,
                'deployments': dict((deployment.name, deployment.representation()) for deployment in self.deployments.itervalues()),
        }
        if self.coordinators:
            spec['coordinators'] = dict((coordinator.id, [coordinator.address, coordinator.port]) for coordinator in self.coordinators.itervalues())
        if self.master_coordinator:
            spec['master_coordinator'] = self.master_coordinator_id
        if self.target_deployment_name:
            spec['target_deployment'] = self.target_deployment_name
        return spec

    def find_nodes_for_key(self, key):
        nodes = dict([(node.id, node) for bucket in self.active_deployment.buckets_for_key(key) for node in bucket])
        if self.target_deployment:
            nodes.update(dict([(node.id, node) for bucket in self.target_deployment.buckets_for_key(key) for node in bucket]))
        return nodes

    def find_neighbour_nodes_for_node(self, local_node):
        neighbour_buckets = []
        if local_node.id in self.active_deployment.nodes:
            deployment_node_bucket = self.active_deployment.buckets.get(local_node.bucket_id, None)
            neighbour_buckets.extend(self.active_deployment.consistent_hash.find_neighbour_buckets(deployment_node_bucket))
        if self.target_deployment and local_node.id in self.target_deployment.nodes:
            target_node_bucket = self.target_deployment.buckets.get(local_node.bucket_id, None)
            neighbour_buckets.extend(self.target_deployment.consistent_hash.find_neighbour_buckets(target_node_bucket))
        neighbour_nodes = dict((node.id, node) for bucket in neighbour_buckets for node in bucket)
        neighbour_nodes.pop(local_node.id, None)
        return neighbour_nodes

    def all_nodes(self):
        nodes = dict((node.id, node) for bucket in self.active_deployment.buckets.itervalues() for node in bucket)
        if self.target_deployment:
            nodes.update([(node.id, node) for bucket in self.target_deployment.buckets.itervalues() for node in bucket])
        return nodes



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

from syncless import coio
from syncless.util import Queue

class Task(object):
    """docstring for Task"""
    def __init__(self, f, *args, **kwargs):
        super(Task, self).__init__()
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.f(*self.args, **self.kwargs)

class Runner(object):
    def __init__(self, tasks):
        super(Runner, self).__init__()
        self.tasks = tasks
        self.completed_count = 0
        self.q = Queue()
        self.results = [None for task in tasks]

    def spawn(self):
        i = 0
        for task in self.tasks:
            assert task
            coio.stackless.tasklet(self._task_runner)(task, i)
            i += 1

    def await_completion(self):
        self.q.pop()

    def _task_runner(self, task, i):
        result = task()
        self._complete(result, i)

    def _complete(self, result, i):
        self.results[i] = result
        self.completed_count += 1
        if self.completed_count == len(self.tasks):
            self.q.append(True)

def run(tasks):
    r = Runner(tasks)
    r.spawn()
    r.await_completion()
    return r.results

def task(func, *args, **kwargs):
    assert func
    return Task(func, *args, **kwargs)

def spawn(func, *args, **kwargs):
    return coio.stackless.tasklet(func)(*args, **kwargs)

if __name__ == '__main__':

    def foo(bar):
        print bar

    run([task(foo, 'hello foo')])



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-
import logging

import paths
paths.setup()

from socketless import service

from utils import debug
from configurationcontroller import ConfigurationController
from protocols import PublicNodeServiceProtocol

class ValueNotAvailableException(BaseException):
    def __init__(self, key):
        super(ValueNotAvailableException, self).__init__('Value not currently available for key "%s"' % repr(key))
        self.key = key

class Client(object):
    def __init__(self, name, coordinator_addresses=[], explicit_configuration=None, configuration_cache_directory=None,
                 configuration_update_interval=5*30):
        super(Client, self).__init__()
        self.__node_clients = {}
        self.__configuration = None
        self.__configuration_controller = ConfigurationController(name, coordinator_addresses, explicit_configuration,
                                                                  configuration_cache_directory, self.__update_configuration,
                                                                  configuration_update_interval)

    def __initialize_node_client_pool(self):
        new_nodes = self.__configuration.all_nodes()
        new_node_clients = {}
        for node_id, client in self.__node_clients.iteritems():
            if node_id in new_nodes:
                new_node_clients[node_id] = client
            else:
                client.close()
        for node_id, node in new_nodes.iteritems():
            if node_id not in new_node_clients:
                new_node_clients[node_id] = service.Client((node.address, node.raw_port), PublicNodeServiceProtocol, tag=node_id)
        self.__node_clients = new_node_clients

    def __update_configuration(self, new_configuration):
        debug.log(new_configuration)
        self.__configuration = new_configuration
        self.__initialize_node_client_pool()

    def __client_for_key(self, key):
        # TODO: Distribute node selection evenly?
        if not self.__configuration:
            return None
        nodes = self.__configuration.find_nodes_for_key(key)
        for node in nodes.itervalues():
            client = self.__node_clients[node.id]
            if client.is_connected():
                return client
        return None

    def connect(self):
        self.__configuration_controller.start()

    def disconnect(self):
        raise Exception('Not implemented')

    def connected_node_count(self):
        return len([node_client for node_client in self.__node_clients.itervalues() if node_client.is_connected()])

    def total_node_count(self):
        return len(self.__node_clients)

    def has_configuration(self):
        return self.__configuration_controller.configuration != None

    def is_connected(self):
        return self.has_configuration() and self.connected_node_count() > self.total_node_count() / 2.0

    def set(self, key, timestamp, value):
        logging.debug('key = %s, timestamp = %s', key, timestamp)
        node_client = self.__client_for_key(key)
        if not node_client:
            raise ValueNotAvailableException(key)
        return node_client.set(key, timestamp, value)

    def get(self, key):
        node_client = self.__client_for_key(key)
        if not node_client:
            raise ValueNotAvailableException(key)
        result = node_client.get(key)
        if not result:
            raise ValueNotAvailableException(key)
        timestamp, value = result
        return timestamp or None, value

    def stat(self, key):
        node_client = self.__client_for_key(key)
        if not node_client:
            raise ValueNotAvailableException(key)
        return node_client.stat(key) or None



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

from datetime import timedelta
import logging

from consistenthash import ConsistentHash
from utils.timedelta_parser import parse_timedelta

def timedelta_total_seconds(td):
    return (td.microseconds + (td.seconds + td.days * 24.0 * 3600.0) * 10**6) / 10**6

def timedelta_days(td):
    return timedelta_total_seconds(td) / (24.0 * 3600.0)

class Node(object):
    """docstring for Node"""
    def __init__(self, node_id, bucket_id, address, http_port, raw_port):
        super(Node, self).__init__()
        self.id = node_id
        self.bucket_id = bucket_id
        self.address = address
        self.http_port = http_port
        self.raw_port = raw_port

    def __str__(self):
        """docstring for __str__"""
        return "Node(id=%s, bucket=%s, http://%s:%d tako://%s:%d)" % (self.id, self.bucket_id, self.address, self.http_port, self.address, self.raw_port)

    def __repr__(self):
        """docstring for __repr__"""
        return str(self)

    def raw_address(self):
        """docstring for raw_address"""
        return (self.address, self.raw_port)

    def store_url(self):
        """docstring for store_url"""
        return 'http://%s:%d/store/' % (self.address, self.http_port)

    def stat_url(self):
        """docstring for stat_url"""
        return 'http://%s:%d/stat/' % (self.address, self.http_port)

class Bucket(object):
    """docstring for Bucket"""
    def __init__(self, bucket_id, nodes):
        super(Bucket, self).__init__()
        self.id = bucket_id
        self.nodes = nodes

    def __str__(self):
        return 'Bucket(%s)' % ', '.join([str(node) for node in self.nodes.itervalues()])

    def __repr__(self):
        return str(self)

    def __iter__(self):
        return self.nodes.itervalues()

    def representation(self):
        return dict((node.id, [node.address, node.http_port, node.raw_port]) for node in self.nodes.itervalues())

    # Essential for consistent hashing, be careful when modifying!
    def __hash__(self):
      return hash(self.id)

class Deployment(object):
    """docstring for Deployment"""
    def __init__(self, name, representation):
        super(Deployment, self).__init__()
        self.original_representation = representation
        self.name = name
        self.buckets = {}
        for bucket_id, bucket in representation['buckets'].iteritems():
            nodes = dict((node_id, Node(node_id, bucket_id, address, http_port, raw_port)) for node_id, (address, http_port, raw_port) in bucket.iteritems())
            self.buckets[bucket_id] = Bucket(bucket_id, nodes)
        self.nodes = dict((node_id, node) for bucket in self.buckets.itervalues() for node_id, node in bucket.nodes.iteritems())
        self.consistent_hash = ConsistentHash(self.buckets.values(), **representation.get('hash', {}))
        self.read_repair_enabled = representation.get('read_repair', True)
        self.background_healing_enabled = representation.get('background_healing', True)
        self.background_healing_interval = parse_timedelta(representation.get('background_healing_interval', '1d'))
        self.background_healing_interval_seconds = timedelta_total_seconds(self.background_healing_interval)
        if self.background_healing_enabled:
            if not self.background_healing_interval_seconds:
                raise Exception('Parsing error')

    def siblings(self, node_id):
        """docstring for siblings"""
        node = self.nodes[node_id]
        bucket = self.buckets[node.bucket_id]
        siblings = dict(bucket.nodes)
        del siblings[node_id]
        return siblings.values()

    def buckets_for_key(self, key):
        return self.consistent_hash.find_buckets(key)

    def representation(self):
        spec = {
                'read_repair': self.read_repair_enabled,
                'background_healing':self.background_healing_enabled,
                'background_healing_interval':str(self.background_healing_interval),
                'hash': {
                        'buckets_per_key': self.consistent_hash.buckets_per_key,
                },
                'buckets':dict((bucket_id, bucket.representation()) for bucket_id, bucket in self.buckets.iteritems()),
        }
        return spec

    def __str__(self):
        """docstring for __str__"""
        return 'Deployment(read_repair=%s, hash=%s, buckets=%s)' % (self.read_repair_enabled, {'buckets_per_key':self.consistent_hash.buckets_per_key}, ', '.join([str(bucket) for bucket in self.buckets.itervalues()]))

    def __repr__(self):
        """docstring for __repr__"""
        return str(self)

class Coordinator(object):
    """docstring for Coordinator"""
    def __init__(self, coordinator_id, address, port):
        super(Coordinator, self).__init__()
        self.id = coordinator_id
        self.address = address
        self.port = port

    def __str__(self):
        """docstring for __str__"""
        return "Coordinator(id=%s, %s:%d)" % (self.id, self.address, self.port)

    def __repr__(self):
        """docstring for __repr__"""
        return str(self)

    def configuration_url(self):
        """docstring for configuration_url"""
        return 'http://%s:%d/configuration' % (self.address, self.port)



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import logging
import os

import paths
paths.setup()

import configuration
from utils import timestamper

MAX_CONFIGURATION_HISTORY = 10

class ConfigurationCache(object):
    """docstring for ConfigurationCache"""
    def __init__(self, directory, name):
        super(ConfigurationCache, self).__init__()
        self.directory = os.path.abspath(directory)
        self.name = name
        logging.debug('name = "%s", directory = "%s"', self.name, self.directory)

    def filename_matches(self, filename):
        name, ext = os.path.splitext(filename)
        if ext != '.yaml':
            return False
        creator_name, timestamp_string = name.split('.')
        return creator_name == self.name

    def list_files(self):
        try:
            return sorted([filename for filename in os.listdir(self.directory) if self.filename_matches(filename)])
        except OSError, e:
            logging.debug(e)
            return []

    def cleanup(self):
        filenames = self.list_files()
        if len(filenames) > MAX_CONFIGURATION_HISTORY:
            remove_count = len(filenames) - MAX_CONFIGURATION_HISTORY
            remove_files = filenames[0:remove_count]
            logging.debug('Removing files: %s' % ', '.join([repr(filename) for filename in remove_files]))
            for filename in remove_files:
                filepath = os.path.join(self.directory, filename)
                try:
                    os.unlink(filepath)
                except OSError, e:
                    logging.error('Failed to remove configuration file: %s', e)

    def get_configuration(self):
        filenames = self.list_files()
        logging.debug('filenames = %s', filenames)
        filenames.reverse()
        for filename in filenames:
            name, ext = os.path.splitext(filename)
            creator_name, timestamp_string = name.split('.')
            timestamp = timestamper.loads(timestamp_string)
            filepath = os.path.join(self.directory, filename)
            persisted_configuration = configuration.try_load_file(filepath, timestamp)
            if persisted_configuration:
                return persisted_configuration
        return None

    def cache_configuration(self, cfg):
        filename = '%s.%s.yaml' % (self.name, cfg.timestamp)
        filepath = os.path.join(self.directory, filename)
        if configuration.try_dump_file(filepath, cfg):
            self.cleanup()
            return True
        return False



import time
import struct

struct_object = struct.Struct('!Q')
pack = struct_object.pack
unpack = lambda s:struct_object.unpack(s)[0]

def dumps(microseconds):
    return '%012d' % microseconds

def loads(s):
    return long(s)

def try_loads(s):
    try:
        return loads(s)
    except ValueError:
        return None
    except TypeError:
        return None


def to_seconds(microseconds):
    return microseconds / 1000000.0

def from_seconds(seconds):
    return long(seconds * 1000000)

def now():
    return from_seconds(time.time())



# -*- Mode: Python; tab-width: 4; indent-tabs-mode: nil; -*-

import urllib3
from urlparse import urlparse
import logging

pool_cache = {}

def _connection_from_urlinfo(urlinfo):
    key = '%s://%s:%s/' % (urlinfo.scheme, urlinfo.hostname, urlinfo.port)
    pool = pool_cache.get(key, None)
    if not pool:
        pool = urllib3.connection_from_url(urlinfo.geturl())
        pool_cache[key] = pool
    return pool

def fetch(url, tries=3):
    """docstring for fetch"""
    urlinfo = urlparse(url)
    pool = _connection_from_urlinfo(urlinfo)
    logging.debug('url: %s, tries: %d', url, tries)
    for i in xrange(1, tries):
        try:
            r = pool.get_url(urlinfo.path)
            return r.data, r.headers
        except IOError, e:
            logging.error('Error: %s', e)
            return (None, None)

def post(url, value, headers={}):
    """docstring for post"""
    urlinfo = urlparse(url)
    pool = _connection_from_urlinfo(urlinfo)
    pool.urlopen('POST', urlinfo.path, body=value, headers=headers)



import re
from datetime import timedelta

def parse_timedelta(s):
    if s is None:
        return None
    d = re.match(
            r'((?P<days>\d+)\s*(days|day|d)\s*,?)?\s*((?P<hours>\d+)\s*:\s*'
            r'(?P<minutes>\d+)\s*:\s*(?P<seconds>\d+))?\s*',
            str(s)).groupdict(0)
    return timedelta(**dict(( (key, int(value))
                              for key, value in d.items() )))

